from typing import Generator, NamedTuple
import requests
import parsel
import datetime
import functools
from .artwork_options import ArtworkOptions, new_filter, is_skip
from .struct import *
from .image_urls import _load_artwork_url_gen, _ArtworkUrl
from .utils import *
import pixivtools.pixiv_cfg as pixiv_cfg


class ArtworkInfo():
    def __init__(self, res: ArtInfoResponseType) -> None:
        self._raw_resp: ArtInfoResponseType = res
        self._resp_json: dict = {}
        self._timezone = datetime.datetime.now().astimezone().tzinfo.tzname(None)

    def _get_body(self) -> dict:
        if self._resp_json:
            return self._resp_json['body']
        if callable(self._raw_resp):
            self._raw_resp = self._raw_resp()._raw_resp
        if isinstance(self._raw_resp, requests.Response):
            self._resp_json = self._raw_resp.json()
        return self._resp_json['body']

    def __str__(self, wrap=False, detail=False):
        self._get_body()
        dct = {}
        if detail:
            dct = self._get_body()
        else:
            dct = {
                "artwork_id": self.artwork_id,
                "user_id": self.user_id,
                "user_name": self.user_name,
                "artwork_type": self.artwork_type.name,
                "tags": [(i.name, i.translation) for i in self.tags],
                "image_download_urls": [str(i) for i in self.image_download_urls],
                "title": self.title,
                "nums": self.nums,
                "restrict": self.restrict.name,
                "desc": self.desc,
                "bookmark_cnt": self.bookmark_cnt,
                "like_cnt": self.like_cnt,
                "comment_cnt": self.comment_cnt,
                "view_cnt": self.view_cnt,
                "create_time": self.create_time.strftime("%Y-%m-%d %H:%M:%S"),
                "upload_time": self.upload_time.strftime("%Y-%m-%d %H:%M:%S"),
                "height": self.height,
                "width": self.width
            }

        if wrap:
            return "\n".join([f"{k}: {v}" for k, v in dct.items()])
        return str(dct)

    def __repr__(self, wrap=False, detail=False):
        return self.__str__(wrap, detail)

    @property
    def artwork_id(self) -> int:
        body = self._get_body()
        return int(body['illustId'])

    @property
    def user_id(self) -> int:
        body = self._get_body()
        return int(body['userId'])

    @property
    def user_name(self) -> str:
        body = self._get_body()
        return body['userName']

    @property
    def artwork_type(self) -> ArtworkType:
        body = self._get_body()
        return ArtworkType(
            int(body['illustType'])
        )

    @property
    def tags(self) -> list[ArtworkTag]:
        body = self._get_body()
        tags: list[ArtworkTag] = []

        used_tag = set()
        for tag_info in body['tags']['tags']:
            tag = tag_info['tag']
            translation = tag_info.get('translation', {})

            tag_trans = translation.get('en', '') if translation else ''

            if tag in used_tag:  # 网站部分返回的会出现重复tag的问题（e.g. 69353795）
                continue
            used_tag.add(tag)
            tags.append(ArtworkTag(name=tag, translation=tag_trans))

        return tags

    @property
    def image_download_urls(self) -> Generator[_ArtworkUrl, None, None]:
        if self.artwork_type == ArtworkType.UGORIA:
            return (_ArtworkUrl() for _ in ())
        body = self._get_body()
        url_gen = _load_artwork_url_gen(body['urls'])
        return (
            url_gen(i)
            for i in range(self.nums)
        )

    @property
    def title(self) -> str:
        body = self._get_body()
        return body['title']

    @property
    def nums(self) -> int:
        body = self._get_body()
        return int(body['pageCount'])

    @property
    def restrict(self) -> ArtworkRestrict:
        body = self._get_body()
        return ArtworkRestrict(
            int(body['xRestrict'])
        )

    @property
    def desc(self) -> str:
        body = self._get_body()
        return body['description']

    @property
    def bookmark_cnt(self) -> int:
        body = self._get_body()
        return body['bookmarkCount']

    @property
    def like_cnt(self) -> int:
        body = self._get_body()
        return body['likeCount']

    @property
    def comment_cnt(self) -> int:
        body = self._get_body()
        return body['commentCount']

    @property
    def view_cnt(self) -> int:
        body = self._get_body()
        return body['viewCount']

    @property
    def create_time(self) -> datetime.datetime:
        body = self._get_body()
        tm = datetime.datetime.fromisoformat(body['createDate'])
        return tm.astimezone(None)

    @property
    def upload_time(self) -> datetime.datetime:
        body = self._get_body()
        tm = datetime.datetime.fromisoformat(body['uploadDate'])
        return tm.astimezone(None)

    @property
    def height(self) -> int:
        body = self._get_body()
        return body['height']

    @property
    def width(self) -> int:
        body = self._get_body()
        return body['width']


class PixivisionInfo(NamedTuple):
    aid: int
    title: str
    desc: str
    pixivision_type: str
    artworks: dict[int, ArtworkInfo]


BASE_HEADERS = {
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, "
                  "like Gecko) Chrome/81.0.4044.138 Safari/537.36",
    "accept-language": "en,zh-CN;q=0.9,zh;q=0.8",
    "referer": "https://www.pixiv.net/",
}


class PixivApi():
    def __init__(self, meta: pixiv_cfg.ApiMetaArgument):
        self._meta = meta
        self._session = requests.session()
        self._init_session()

    def _init_session(self):
        if self._meta.proxy:
            self._session.proxies.update({
                "http": self._meta.proxy,
                "https": self._meta.proxy
            })
        if self._meta.phpsessid:
            self._session.cookies.update({
                "PHPSESSID": self._meta.phpsessid
            })

    def _gen_artwork_info_dict(self, artwork_ids: list[int], options: ArtworkOptions) -> dict[int, ArtworkInfo]:
        return {
            i: ArtworkInfo(functools.partial(
                self.get_artwork_info, artwork_id=i, options=options
            ))
            for i in artwork_ids
        }

    def get_artwork_info(self, artwork_id: int, options: ArtworkOptions) -> ArtworkInfo:
        # 根据artwork_id获取artwork信息
        url = f"https://www.pixiv.net/ajax/illust/{artwork_id}?lang=zh"
        res = self._session.get(url=url, headers=BASE_HEADERS)
        return ArtworkInfo(res)

    def get_artworks_by_userid(self, user_id: int, options: ArtworkOptions) -> dict[int, ArtworkInfo]:
        # 根据user_id获取该用户的所有artwork信息
        url = f"https://www.pixiv.net/ajax/user/{user_id}/profile/all?lang=zh"
        res = self._session.get(url=url, headers=BASE_HEADERS).json()
        artwork_ids = res['body']['illusts']
        artwork_ids = get_unique_ids([int(i) for i in artwork_ids])
        return self._gen_artwork_info_dict(artwork_ids, options)

    def get_artworks_by_follow_latest(self, page: int, options: ArtworkOptions) -> dict[int, ArtworkInfo]:
        # 根据我的账号获取关注用户的最新的artwork信息
        url = f"https://www.pixiv.net/ajax/follow_latest/illust?p={page}&lang=zh"
        url += "&mode=r18" if options['only_r18'] else "&mode=all"
        res = self._session.get(url=url, headers=BASE_HEADERS).json()
        artwork_ids = res['body']['page']['ids']
        artwork_ids = get_unique_ids([int(i) for i in artwork_ids])
        return self._gen_artwork_info_dict(artwork_ids, options)

    def get_artworks_by_pixivision_aid(self, aid: int, options: ArtworkOptions) -> PixivisionInfo:
        # 根据pixivision的aid获取pixivision
        url = f"https://www.pixivision.net/zh/a/{aid}"
        headers = {
            **BASE_HEADERS,
            "Accept-Language": "zh-CN,zh;q=0.9,en;q=0.8"
        }
        res = self._session.get(url=url, headers=headers).text

        parsel_obj = parsel.Selector(res)
        title = parsel_obj.xpath('//meta[@property="og:title"]/@content').get()
        a_type = parsel_obj.css('.am__categoty-pr').css('a').attrib['data-gtm-label']
        description = parsel_obj.xpath('//meta[@property="og:description"]/@content').get()

        if a_type != 'illustration':
            return PixivisionInfo(aid, title, description, a_type, {})

        artwork_ids = []
        for i in parsel_obj.css('.am__body')[0].css('.am__work__main'):
            href = i.css('a').attrib['href']
            artwork_id = href.split('/')[-1]
            if '?' in artwork_id:
                artwork_id = artwork_id.split('?')[0]
            artwork_ids.append(int(artwork_id))

        res = PixivisionInfo(
            aid, title, description, a_type,
            self._gen_artwork_info_dict(artwork_ids, options)
        )
        return res

    def get_image(self, url: str) -> bytes:
        res = self._session.get(url=url, headers=BASE_HEADERS)
        res.raise_for_status()
        return res.content

    def get_artworks_by_recommend(self, options: ArtworkOptions) -> dict[int, ArtworkInfo]:
        # 根据我的账号获取推荐的artwork信息
        url = "https://www.pixiv.net/ajax/top/illust?lang=zh"
        url += "&mode=r18" if options['only_r18'] else "&mode=all"
        res = self._session.get(url=url, headers=BASE_HEADERS).json()
        artwork_ids = res['body']['page']['recommend']['ids']
        artwork_ids = get_unique_ids(int(i) for i in artwork_ids)
        return self._gen_artwork_info_dict(artwork_ids, options)

    def get_artworks_by_rank(self, rank_type: RankType, date: int, page: int, options: ArtworkOptions) -> dict[int, ArtworkInfo]:
        # 根据rank_type获取rank的artwork信息
        # date: 日期，格式为YYYYMMDD
        url = f"https://www.pixiv.net/ranking.php?&content=illust&p=1&format=json"
        url += f"&date={date}&mode={rank_type.value}"
        res = self._session.get(url=url, headers=BASE_HEADERS).json()
        artwork_ids = [int(i['illust_id']) for i in res['contents']]
        artwork_ids = get_unique_ids(int(i) for i in artwork_ids)
        return self._gen_artwork_info_dict(artwork_ids, options)

    def get_artworks_by_request_recommend(self, options: ArtworkOptions) -> dict[int, ArtworkInfo]:
        # 获取约稿作品推荐的artwork信息
        url = f"https://www.pixiv.net/ajax/commission/page/request/complete/illust?p=1&lang=zh"
        if options['only_r18']:
            url += "&mode=r18"
        elif options['only_non_r18']:
            url += "&mode=safe"
        res = self._session.get(url=url, headers=BASE_HEADERS).json()
        reqs = res['body']['requests']
        artwork_ids = [int(req['postWork']['postWorkId']) for req in reqs]
        artwork_ids = get_unique_ids(artwork_ids)
        return self._gen_artwork_info_dict(artwork_ids, options)

    def get_userids_by_request_creator(self, options: ArtworkOptions) -> list[int]:
        # 获取约稿作品推荐的用户id
        url = f"https://www.pixiv.net/ajax/commission/page/request/creators/illust/ids?&follows=0&p=1&lang=zh"
        if options['only_r18']:
            url += "&mode=r18"
        elif options['only_non_r18']:
            url += "&mode=safe"
        res = self._session.get(url=url, headers=BASE_HEADERS).json()
        userids = res['body']['page']['creatorUserIds']
        userids = get_unique_ids([int(creator) for creator in userids])
        return userids

    def get_userids_by_similar_user(self, user_id: int, options: ArtworkOptions) -> list[int]:
        # 根据用户id获取相似的用户id
        url = f"https://www.pixiv.net/ajax/user/{user_id}/recommends?userNum=20&workNum=3&lang=zh"
        url += "&isR18=false" if options['only_non_r18'] else "&isR18=true"
        res = self._session.get(url=url, headers=BASE_HEADERS).json()
        userids = [int(i['userId']) for i in res['body']['recommendUsers']]
        userids = get_unique_ids(userids)
        return userids

    def get_artworks_by_user_bookmark(self, user_id: int, page: int, options: ArtworkOptions) -> dict[int, ArtworkInfo]:
        # 根据用户id获取收藏的artwork信息
        url = f"https://www.pixiv.net/ajax/user/{user_id}/illusts/bookmarks?tag=&offset={(page-1)*48}&limit=48&rest=show&lang=zh"
        res = self._session.get(url=url, headers=BASE_HEADERS).json()
        artwork_ids = [int(i['id']) for i in res['body']['works']]
        artwork_ids = get_unique_ids(artwork_ids)
        return self._gen_artwork_info_dict(artwork_ids, options)

    def get_artworks_by_tag_popular(self, tag_name: str, options: ArtworkOptions) -> dict[int, ArtworkInfo]:
        # 根据tag获取热门的artwork信息
        tag_name = requests.utils.quote(tag_name)
        url = f"https://www.pixiv.net/ajax/search/top/{tag_name}?lang=zh"
        res = self._session.get(url=url, headers=BASE_HEADERS).json()
        populars = res['body']['popular']
        artworks = populars['permanent'] + populars['recent']
        artwork_ids = [int(i["id"]) for i in artworks]
        artwork_ids = get_unique_ids(artwork_ids)
        return self._gen_artwork_info_dict(artwork_ids, options)

    def get_userids_by_recommend(self, options: ArtworkOptions) -> list[int]:
        # 根据我的账号获取推荐的用户id
        url = "https://www.pixiv.net/ajax/top/illust?mode=all&lang=zh"
        res = self._session.get(url=url, headers=BASE_HEADERS).json()
        userids = [int(i['userId']) for i in res['body']['users']]
        userids = get_unique_ids(userids)
        return userids

    def get_artworks_by_similar_artwork(self, artwork_id: int, options: ArtworkOptions) -> dict[int, ArtworkInfo]:
        # 根据artwork_id获取相似的artwork信息
        url = f"https://www.pixiv.net/ajax/illust/{artwork_id}/recommend/init?limit=20&lang=zh"
        res = self._session.get(url=url, headers=BASE_HEADERS).json()
        artwork_ids = [int(i['id']) for i in res['body']['illusts']]
        artwork_ids = get_unique_ids(artwork_ids)
        return self._gen_artwork_info_dict(artwork_ids, options)


def new_pixiv_api(meta: pixiv_cfg.ApiMetaArgument) -> PixivApi:
    return PixivApi(meta)



from typing import Optional, TypedDict
from .struct import *


class ArtworkOptions(TypedDict):
    # update: bool, 是否更新已存在的artwork
    update: bool
    # only_r18: bool, 是否只爬取R18
    only_r18: bool
    # only_non_r18: bool, 是否只爬取非R18
    only_non_r18: bool
    # skip_manga: bool, 是否跳过漫画
    skip_manga: bool
    # artwork_types: list[ArtworkType], 只爬取指定类型的artwork
    artwork_types: list[ArtworkType] | None
    # ignore_error: bool, 是否忽略爬取过程中的某个artwork出错，如果为False则会在出错时直接raise
    ignore_error: bool


def is_skip(options: ArtworkOptions, artwork_info) -> Optional[str]:
    if options['only_r18'] and artwork_info.restrict == ArtworkRestrict.NON_R18:
        return "only_r18"
    if options['only_non_r18'] and artwork_info.restrict != ArtworkRestrict.NON_R18:
        return "only_non_r18"
    if options['skip_manga'] and artwork_info.artwork_type == ArtworkType.MANGA:
        return "skip_manga"
    if options['artwork_types'] and artwork_info.artwork_type not in options['artwork_types']:
        return "artwork_types"


def _default_filter() -> ArtworkOptions:
    return ArtworkOptions(
        update=False,
        only_r18=False,
        only_non_r18=False,
        skip_manga=True,
        artwork_types=None,
        ignore_error=True
    )


def new_filter(**kwargs: ArtworkOptions) -> ArtworkOptions:
    res = _default_filter()
    res.update(kwargs)
    return res



from .crawler import PixivCrawler
from .context import new_context
import pixivtools.pixiv_cfg as cfg


class PixivService():
    def __init__(self, cfg: cfg.PixivConfig):
        self._context = new_context(cfg)
        self._crawler = None

    def crawler(self) -> PixivCrawler:
        if self._crawler:
            return self._crawler
        self._crawler = PixivCrawler(self._context)
        return self._crawler
    
    def sql(self):
        return self._context.sql
    
    def api(self):
        return self._context.api


def new_pixiv_service(cfg: cfg.PixivConfig) -> PixivService:
    return PixivService(cfg)



from typing import Callable, NamedTuple
import enum
import requests


LazyResponse = Callable[[], requests.Response]
ArtInfoResponseType = requests.Response | LazyResponse


class ArtworkType(enum.Enum):
    ILLUST = 0  # 插画
    MANGA = 1  # 漫画
    UGORIA = 2  # 动图


class ArtworkTag(NamedTuple):
    name: str
    translation: str


class ArtworkRestrict(enum.Enum):
    NON_R18 = 0
    R18 = 1
    R18G = 2


class RankType(enum.Enum):
    DAILY = 'daily'  # 日榜
    WEEKLY = 'weekly'  # 周榜
    MONTHLY = 'monthly'  # 月榜
    NEWBIE = "rookie"  # 新人榜
    ORIGINAL = "original"  # 原创榜
    DAILY_AI = "daily_ai"  # 日榜-ai
    MALE = "male"  # 男性喜爱
    FEMALE = "female"  # 女性喜爱

    DAILY_R18 = 'daily_r18'
    WEEKLY_R18 = 'weekly_r18'
    WEEKLY_R18G = 'r18g'
    DAILY_AI_R18 = "daily_r18_ai"
    MALE_R18 = "male_r18"
    FEMALE_R18 = "female_r18"



import unittest
from pixivtools.pixiv_api import PixivApi, new_filter, RankType
import pixivtools.pixiv_cfg as pixiv_cfg
import datetime


TEST_ARTWORK_IDS = [128168801]  # 使用有效的artwork_id进行测试
TEST_USER_IDS = [85966]         # 使用有效的user_id进行测试
TEST_PIXIVISION_AIDS = [10176]  # 使用有效的aid进行测试
TEST_TAGS = ["東方Project"]     # 使用有效的tag进行测试


class TestPixivApi(unittest.TestCase):
    def setUp(self):
        cfg = pixiv_cfg.get_pixiv_config()
        self.api = PixivApi(cfg.api)
        self.options = new_filter()

    def test_get_artwork_info(self):
        for artwork_id in TEST_ARTWORK_IDS:
            artwork_info = self.api.get_artwork_info(artwork_id, self.options)
            #print(artwork_info.__str__(wrap=True, detail=False))
            self.assertIsNotNone(artwork_info)
            self.assertEqual(artwork_info.artwork_id, artwork_id)

    def test_get_artworks_by_userid(self):
        for user_id in TEST_USER_IDS:
            artworks = self.api.get_artworks_by_userid(user_id, self.options)
            #print(artworks)
            self.assertIsInstance(artworks, dict)
            self.assertGreater(len(artworks), 0)

    def test_get_artworks_by_follow_latest(self):
        for page in range(1, 3):
            artworks = self.api.get_artworks_by_follow_latest(page, self.options)
            #print(artworks)
            self.assertIsInstance(artworks, dict)
            self.assertGreater(len(artworks), 0)

    def test_get_artworks_by_pixivision_aid(self):
        for aid in TEST_PIXIVISION_AIDS:
            pixivision_info = self.api.get_artworks_by_pixivision_aid(aid, self.options)
            #print(pixivision_info)
            self.assertIsNotNone(pixivision_info)
            self.assertEqual(pixivision_info.aid, aid)

    def test_get_image(self):
        for artwork_id in TEST_ARTWORK_IDS:
            artwork_info = self.api.get_artwork_info(artwork_id, self.options)
            test_imgs = list(artwork_info.image_download_urls)
            if len(test_imgs) > 1:
                test_imgs = [test_imgs[0], test_imgs[-1]]
            for url in test_imgs:
                image_data = self.api.get_image(url.small)
                #print(f"url: {url.small}, image_len: {len(image_data)}")
                self.assertIsInstance(image_data, bytes)
                self.assertGreater(len(image_data), 0)

    def test_get_artworks_by_recommend(self):
        artworks = self.api.get_artworks_by_recommend(self.options)
        #print(artworks)
        self.assertIsInstance(artworks, dict)
        self.assertGreater(len(artworks), 0)

    def test_get_artworks_by_rank(self):
        test_lst = [RankType.DAILY, RankType.WEEKLY_R18, RankType.WEEKLY]
        yesterday = (datetime.datetime.now() - datetime.timedelta(days=1)).strftime("%Y%m%d")
        #print(yesterday)
        for rank_type in test_lst:
            artworks = self.api.get_artworks_by_rank(rank_type, yesterday, 1, self.options)
            #print(artworks)
            self.assertIsInstance(artworks, dict)
            self.assertGreater(len(artworks), 0)

    def test_get_artworks_by_request_recommend(self):
        artworks = self.api.get_artworks_by_request_recommend(self.options)
        #print(artworks)
        self.assertIsInstance(artworks, dict)
        self.assertGreater(len(artworks), 0)
    
    def test_get_userids_by_request_creator(self):
        user_ids = self.api.get_userids_by_request_creator(self.options)
        #print(user_ids)
        self.assertIsInstance(user_ids, list)
        self.assertGreater(len(user_ids), 0)

    def test_get_userids_by_similar_user(self):
        for user_id in TEST_USER_IDS:
            user_ids = self.api.get_userids_by_similar_user(user_id, self.options)
            #print(user_ids)
            self.assertIsInstance(user_ids, list)
            self.assertGreater(len(user_ids), 0)

    def test_get_artworks_by_user_bookmark(self):
        for user_id in TEST_USER_IDS:
            artworks = self.api.get_artworks_by_user_bookmark(user_id, 1, self.options)
            #print(artworks)
            self.assertIsInstance(artworks, dict)
            self.assertGreater(len(artworks), 0)

    def test_get_artworks_by_tag_popular(self):
        for tag in TEST_TAGS:
            artworks = self.api.get_artworks_by_tag_popular(tag, self.options)
            #print(artworks)
            self.assertIsInstance(artworks, dict)
            self.assertGreater(len(artworks), 0)

    def test_get_userids_by_recommend(self):
        user_ids = self.api.get_userids_by_recommend(self.options)
        #print(user_ids)
        self.assertIsInstance(user_ids, list)
        self.assertGreater(len(user_ids), 0)

    def test_get_artworks_by_similar_artwork(self):
        for artwork_id in TEST_ARTWORK_IDS:
            artworks = self.api.get_artworks_by_similar_artwork(artwork_id, self.options)
            #print(artworks)
            self.assertIsInstance(artworks, dict)
            self.assertGreater(len(artworks), 0)

if __name__ == '__main__':
    unittest.main()



import pathlib
from typing import NamedTuple


class ApiMetaArgument(NamedTuple):
    phpsessid: str
    proxy: str


class CrawlerConfig(NamedTuple):
    img_dir: pathlib.Path
    sql_url: str


class PixivConfig(NamedTuple):
    api: ApiMetaArgument
    crawler: CrawlerConfig
    log_file: pathlib.Path


def _load_api_config(api_obj: dict) -> ApiMetaArgument:
    return ApiMetaArgument(**api_obj)


def _load_crawler_config(crawler_obj: dict) -> CrawlerConfig:
    img_file_dir = pathlib.Path() / crawler_obj["img_dir"]
    if not img_file_dir.exists():
        img_file_dir.mkdir(parents=True)

    crawler_obj["img_dir"] = img_file_dir
    return CrawlerConfig(**crawler_obj)


def _load_config(obj: dict) -> PixivConfig:
    log_file_path = pathlib.Path() / obj["log_file"]
    log_file_dir = log_file_path.parent
    if not log_file_dir.exists():
        log_file_dir.mkdir(parents=True)

    return PixivConfig(
        api=_load_api_config(obj["api"]),
        crawler=_load_crawler_config(obj["crawler"]),
        log_file=log_file_path
    )


g_cfg = None


def get_pixiv_config(filename: str = "config.yaml") -> PixivConfig:
    global g_cfg
    if g_cfg:
        return g_cfg

    import yaml
    cfg_path = pathlib.Path() / filename
    if not cfg_path.exists():
        raise FileNotFoundError(f"文件{cfg_path.absolute}不存在")
    with open(cfg_path, "r") as f:
        obj = yaml.safe_load(f)

    g_cfg = _load_config(obj)
    return g_cfg



from typing import NamedTuple, Callable


class _ArtworkUrl:
    def __init__(self, _template: str, i: int):
        self._template = _template
        self._idx = i

    @property
    def original(self): # 原图
        return self._template['original'].format(self._idx)
    
    @property
    def regular(self):  # 一般(一般是1200*1200) https://i.pximg.net/img-master/img/
        return self._template['regular'].format(self._idx)
    
    @property
    def small(self):    # 小图(一般是540*540) https://i.pximg.net/c/48x48/custom-thumb/img/...
        return self._template['small'].format(self._idx)
    
    @property
    def thumb(self):    # 缩略图(一般250*250) https://i.pximg.net/c/250x250_80_a2/custom-thumb/img/...
        return self._template['thumb'].format(self._idx)
    
    @property
    def mini(self):     # 迷你图(一般48*48) https://i.pximg.net/c/48x48/custom-thumb/img/...
        return self._template['mini'].format(self._idx)

    def __str__(self):
        return f"original: {self.original}, regular: {self.regular}, small: {self.small}, thumb: {self.thumb}, mini: {self.mini}"


key2transfunc = {
    'original': lambda x: '_p{}.'.join(x.split('_p0.')),
    'regular': lambda x: '_p{}_'.join(x.split('_p0_')),
    'small': lambda x: '_p{}_'.join(x.split('_p0_')),
    'thumb': lambda x: '_p{}_'.join(x.split('_p0_')),
    'mini': lambda x: '_p{}_'.join(x.split('_p0_')),
}


def _load_artwork_url_gen(urls) -> Callable[[int], _ArtworkUrl]:
    inject = {}
    for k, v in urls.items():
        inject[k] = key2transfunc[k](v)
    def gen_url(idx):
        return _ArtworkUrl(inject, idx)
    return gen_url



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="pixivtools",
    version="0.0.2",
    author="aliubo",
    author_email="liubo@aliubo.com",
    description="A Python module for Pixiv crawling and interaction",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aliubo/pixivtools",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "requests",
        "parsel",
        "pillow",
        "pyyaml",
        "sqlalchemy",
    ],
)



import pixivtools.pixiv_api as pixiv_api
from . import model
from .context import PixivContext
from pathlib import Path
from PIL import Image
import io
from typing import Sequence


class PixivCrawler:
    def __init__(self, context: PixivContext):
        self._api = context.api
        self._sql = context.sql
        self._img_dir = context.img_dir
        self._log = context.log

    def _new_artwork(
            self,
            artwork_info: pixiv_api.ArtworkInfo,
            tags: Sequence[model.Tag],
    ) -> model.Artwork:
        return model.Artwork(
            artwork_id=artwork_info.artwork_id,
            user_id=artwork_info.user_id,
            artwork_type=artwork_info.artwork_type.value,
            title=artwork_info.title,
            nums=artwork_info.nums,
            restrict=artwork_info.restrict.value,
            description=artwork_info.desc,
            bookmark_cnt=artwork_info.bookmark_cnt,
            like_cnt=artwork_info.like_cnt,
            comment_cnt=artwork_info.comment_cnt,
            view_cnt=artwork_info.view_cnt,
            create_time=artwork_info.create_time,
            upload_time=artwork_info.upload_time,
            tags=tags,
        )

    def _get_filepath(self, artwork_id: int, idx: int) -> Path:
        return self._img_dir / f"{artwork_id}_{idx}.jpg"

    def _download_artwork(self, artwork_info: pixiv_api.ArtworkInfo) -> list[model.Image]:
        res = []
        for idx, url in enumerate(artwork_info.image_download_urls):
            file_path = self._get_filepath(artwork_info.artwork_id, idx)
            if file_path.exists() and file_path.stat().st_size > 0:
                with Image.open(file_path) as img:
                    width, height = img.size
                res.append(model.Image(
                    artwork_id=artwork_info.artwork_id,
                    idx=idx,
                    height=height,
                    width=width,
                    file_size=file_path.stat().st_size
                ))
                continue
            content = self._api.get_image(url.original)
            with Image.open(io.BytesIO(content)) as img:
                width, height = img.size
            res.append(model.Image(
                artwork_id=artwork_info.artwork_id,
                idx=idx,
                height=height,
                width=width,
                file_size=len(content)
            ))
            with file_path.open('wb+') as f:
                f.write(content)
        return res

    def _is_artwork_exist(self, artwork_id: int) -> bool:
        with self._sql() as session:
            artwork_record = session.query(model.Artwork).filter_by(artwork_id=artwork_id).first()
            if not artwork_record:
                return False
        for idx in range(artwork_record.nums):
            file_path = self._get_filepath(artwork_id, idx)
            if not file_path.exists() or file_path.stat().st_size <= 0:
                return False
        return True

    def _get_by_artwork_info(
            self,
            artwork_info: pixiv_api.ArtworkInfo,
            options: pixiv_api.ArtworkOptions | None = None
    ) -> bool:
        if options is None:
            options = pixiv_api.new_filter()

        invalid_reason = pixiv_api.is_skip(options, artwork_info)
        if invalid_reason:
            self._log.info(f"artwork {artwork_info.artwork_id} is invalid, reason: {invalid_reason}")
            return False

        if self._is_artwork_exist(artwork_info.artwork_id):
            if not options["update"]:
                self._log.info(f"artwork {artwork_info.artwork_id} already exist")
                return False
        # 爬取图片
        add_images = self._download_artwork(artwork_info)
        # 存数据库
        with self._sql() as session:
            tags = []
            imgs = []
            for tag in artwork_info.tags:
                tag_record = session.query(model.Tag).filter_by(name=tag.name).first()
                if tag_record is None:
                    tags.append(model.Tag(name=tag.name, trans_name=tag.translation))
                else:
                    tags.append(tag_record)
            user = model.User(user_id=artwork_info.user_id, user_name=artwork_info.user_name)
            session.merge(user)
            artwork = self._new_artwork(artwork_info, tags)
            session.merge(artwork)  # 会自动插入不存在的tag
            for img in add_images:
                img_record = session.query(model.Image).filter_by(artwork_id=img.artwork_id, idx=img.idx).first()
                if img_record is None:
                    imgs.append(img)
                else:
                    imgs.append(img_record)
            session.add_all(imgs)

            session.commit()
        return True

    def _get_by_artworks_info(
            self,
            artworks_info: dict[int, pixiv_api.ArtworkInfo],
            options: pixiv_api.ArtworkOptions) -> list[int]:
        self._log.info("Artworks start downloading...", artworks=artworks_info.keys())
        ok_ids = []
        for idx, (artwork_id, artwork_info) in enumerate(artworks_info.items()):
            self._log.info(f"{idx+1}/{len(artworks_info)} - {artwork_id}")
            try:
                if self._get_by_artwork_info(artwork_info, options):
                    ok_ids.append(artwork_id)
            except Exception as e:
                self._log.error(f"save artwork {artwork_id} failed", error=str(e))
                if not options["ignore_error"]:
                    raise e
                continue
            self._log.info(
                f"save artwork {artwork_id} to database",
                title=artwork_info.title,
                tags=[tag.name for tag in artwork_info.tags],
                user=f"{artwork_info.user_name}({artwork_info.user_id})",
                nums=artwork_info.nums
            )
        self._log.info("Artworks download finished", failed_ids=[i for i in artworks_info.keys() if i not in ok_ids])
        return ok_ids

    def get_by_artwork_id(self, artwork_id: int, options: pixiv_api.ArtworkOptions | None = None):
        if options is None:
            options = pixiv_api.new_filter()

        artwork_info = self._api.get_artwork_info(artwork_id, options)
        self._get_by_artwork_info(artwork_info, options)
        self._log.info(
            f"save artwork {artwork_info.artwork_id} to database",
            title=artwork_info.title,
            tags=[tag.name for tag in artwork_info.tags],
            user=f"{artwork_info.user_name}({artwork_info.user_id})",
        )

    def get_by_user_id(self, user_id: int, options: pixiv_api.ArtworkOptions | None = None):
        if options is None:
            options = pixiv_api.new_filter()
        artworks = self._api.get_artworks_by_userid(user_id, options)
        self._log.info("get artworks from user", user_id=user_id)
        self._get_by_artworks_info(artworks, options)

    def _get_by_users_id(self, user_ids: list[int], options: pixiv_api.ArtworkOptions):
        self._log.info("Users start downloading...", user_ids=user_ids)
        for idx, user_id in enumerate(user_ids):
            self._log.info(f"! {idx + 1}/{len(user_ids)} start to save user {user_id} to database")
            self.get_by_user_id(user_id, options)
        self._log.info("Users download finished")

    def get_by_pixivision_aid(self, aid: int, options: pixiv_api.ArtworkOptions | None = None):
        if options is None:
            options = pixiv_api.new_filter()

        res = self._api.get_artworks_by_pixivision_aid(aid, options)
        self._log.info(f"get artworks from pixivision", aid=aid, title=res.title, type=res.pixivision_type)
        get_ids = self._get_by_artworks_info(res.artworks, options)
        with self._sql() as session:
            pixivision = model.Pixivision(
                aid=aid,
                title=res.title,
                type=res.pixivision_type,
                description=res.desc,
                artworks=[
                    session.query(model.Artwork).filter_by(artwork_id=artwork_id).first()
                    for artwork_id in get_ids
                ]
            )
            session.merge(pixivision)
            session.commit()
        self._log.info(f"save pixivision {aid} to database")

    def get_by_follow_latest(self, page: int, options: pixiv_api.ArtworkOptions | None = None):
        if options is None:
            options = pixiv_api.new_filter()

        artworks = self._api.get_artworks_by_follow_latest(page, options)
        self._log.info(f"get artworks from bookmark new", page=page)
        self._get_by_artworks_info(artworks, options)

    def get_by_recommend(self, options: pixiv_api.ArtworkOptions | None = None):
        if options is None:
            options = pixiv_api.new_filter()

        artworks: dict[int, pixiv_api.ArtworkInfo] = self._api.get_artworks_by_recommend(options)
        self._log.info(f"get artworks from recommend")
        self._get_by_artworks_info(artworks, options)

    def get_by_rank(self, rank_type: pixiv_api.RankType, date: int, page: int, options: pixiv_api.ArtworkOptions | None = None):
        if options is None:
            options = pixiv_api.new_filter()

        artworks: dict[int, pixiv_api.ArtworkInfo] = self._api.get_artworks_by_rank(rank_type, date, page, options)
        self._log.info(f"get artworks from rank", rank_type=rank_type.name, date=date, page=page)
        self._get_by_artworks_info(artworks, options)

    def get_by_request_recommend(self, options: pixiv_api.ArtworkOptions | None = None):
        if options is None:
            options = pixiv_api.new_filter()

        artworks: dict[int, pixiv_api.ArtworkInfo] = self._api.get_artworks_by_request_recommend(options)
        self._log.info(f"get artworks from request recommend")
        self._get_by_artworks_info(artworks, options)

    def get_by_user_bookmark(self, user_id: int, page: int, options: pixiv_api.ArtworkOptions | None = None):
        if options is None:
            options = pixiv_api.new_filter()

        artworks: dict[int, pixiv_api.ArtworkInfo] = self._api.get_artworks_by_user_bookmark(user_id, page, options)
        self._log.info(f"get artworks from user bookmark", user_id=user_id, page=page)
        self._get_by_artworks_info(artworks, options)

    def get_by_tag_popular(self, tag_name: str, options: pixiv_api.ArtworkOptions | None = None):
        if options is None:
            options = pixiv_api.new_filter()

        artworks: dict[int, pixiv_api.ArtworkInfo] = self._api.get_artworks_by_tag_popular(tag_name, options)
        self._log.info(f"get artworks from tag popular", tag_name=tag_name)
        self._get_by_artworks_info(artworks, options)

    def get_by_similar_artwork(self, artwork_id: int, options: pixiv_api.ArtworkOptions = None):
        if options is None:
            options = pixiv_api.new_filter()

        artworks: dict[int, pixiv_api.ArtworkInfo] = self._api.get_artworks_by_similar_artwork(artwork_id, options)
        self._log.info(f"get artworks from similar artwork_info", artwork_id=artwork_id)
        self._get_by_artworks_info(artworks, options)

    def get_by_similar_user(self, user_id: int, options: pixiv_api.ArtworkOptions | None = None):
        if options is None:
            options = pixiv_api.new_filter()

        userids: list[int] = self._api.get_userids_by_similar_user(user_id, options)
        self._log.info(f"get similar user from user {user_id}", user_id=user_id)
        self._get_by_users_id(userids, options)

    def get_by_recommend_user(self, options: pixiv_api.ArtworkOptions | None = None):
        if options is None:
            options = pixiv_api.new_filter()

        userids: list[int] = self._api.get_userids_by_recommend(options)
        self._log.info(f"get recommend user")
        self._get_by_users_id(userids, options)

    def get_by_request_creator(self, options: pixiv_api.ArtworkOptions | None = None):
        if options is None:
            options = pixiv_api.new_filter()

        userids: list[int] = self._api.get_userids_by_request_creator(options)
        self._log.info(f"get request creator")
        self._get_by_users_id(userids, options)
